from .transform import with_geo, with_bbl, to_csv, extract_csvs_from_zip, skip_fields
from .transform import hpd_registrations_address_cleanup, hpd_contacts_address_cleanup
from .dof_parser import parse_dof_file


def ecb_violations(dataset):
    return with_bbl(to_csv(dataset.files[0].dest), borough='boro')


def dob_violations(dataset):
    return with_bbl(to_csv(dataset.files[0].dest), borough='boro')


def pluto_16v2(dataset):
    return with_geo(to_csv(extract_csvs_from_zip(dataset.files[0].dest)))


def pluto_17v1(dataset):
    return with_geo(to_csv(extract_csvs_from_zip(dataset.files[0].dest)))


def pluto_18v1(dataset):
    return with_geo(to_csv(extract_csvs_from_zip(dataset.files[0].dest)))


def hpd_complaints(dataset):
    return with_bbl(to_csv(dataset.files[0].dest))


def hpd_complaint_problems(dataset):
    return to_csv(dataset.files[1].dest)


def dob_complaints(dataset):
    return to_csv(dataset.files[0].dest)


def hpd_violations(dataset):
    return to_csv(dataset.files[0].dest)


def hpd_registrations(dataset):
    return hpd_registrations_address_cleanup(to_csv(dataset.files[0].dest))


def hpd_contacts(dataset):
    return hpd_contacts_address_cleanup(to_csv(dataset.files[1].dest))


def dof_sales(dataset):
    for f in dataset.files:
        for row in with_bbl(parse_dof_file(f.dest)):
            yield row


def dobjobs(dataset):
    return with_bbl(to_csv(dataset.files[0].dest))


def rentstab(dataset):
    return to_csv(dataset.files[0].dest)


def rentstab_summary(dataset):
    return to_csv(dataset.files[0].dest)


def acris(dataset, schema):
    dest_file = next(filter(lambda f: schema['table_name'] in f.dest, dataset.files))
    _to_csv = to_csv(dest_file.dest)
    if 'skip' in schema:
        return skip_fields(_to_csv, [s.lower() for s in schema['skip']])
    else:
        return _to_csv


def marshal_evictions_17(dataset):
    return to_csv(dataset.files[0].dest)


def oath_hearings(dataset):
    return with_bbl(to_csv(dataset.files[0].dest),
                    borough='violationlocationborough',
                    block='violationlocationblockno',
                    lot='violationlocationlotno')



import datetime
import logging
import itertools
import os
import subprocess
from functools import lru_cache
from tqdm import tqdm

from . import verify
from . import dataset_transformations
from . import sql
from .database import Database
from .typecast import Typecast
from .file import File
from .utility import read_yml, list_wrap, merge

BATCH_SIZE = 1000


@lru_cache()
def datasets():
    """Returns a dictionary with all defined datasets"""
    return read_yml(os.path.join(os.path.dirname(__file__), 'datasets.yml'))


class Dataset:
    """Information about a dataset"""

    def __init__(self, dataset_name, args=None):
        self.name = dataset_name
        self.args = args
        self.db = None

        if self.args:
            self.root_dir = self.args.root_dir
        else:
            self.root_dir = './data'

        self.dataset = datasets()[dataset_name]
        self.files = self._files()
        self.schemas = list_wrap(self.dataset['schema'])

    def _files(self):
        return [File(file_dict, folder=self.name, root_dir=self.root_dir) for file_dict in self.dataset['files']]


    def download_files(self):
        for f in self.files:
            f.download()


    def db_import(self):
        """
        inserts the dataset in the postgres
        output:  True | Throws
        """
        self.setup_db()
        self.create_schema()

        for schema in self.schemas:
            self.import_schema(schema)

        self.sql_files()

    def index(self):
        if 'index' in self.dataset:
            for sql_file in self.dataset['index']:
                self.db.execute_sql_file(sql_file)
        else:
            logging.debug('no index files exist for this dataset')

    def transform(self, schema):
        """
        Calls the function in dataset_transformation with the same name
        as the schema.

        If no function exists with the same name as the schema table, it tries
        to call the function with the same name as the dataset

        input: dict
        output: generator
        """
        tc = Typecast(schema)

        try:
            rows = getattr(dataset_transformations, schema['table_name'])(self)
        except AttributeError:
            rows = getattr(dataset_transformations, self.name)(self, schema)

        return tc.cast_rows(rows)

    def import_schema(self, schema):
        rows = self.transform(schema)

        pbar = tqdm(unit='rows')
        while True:
            batch = list(itertools.islice(rows, 0, BATCH_SIZE))
            if len(batch) == 0:
                break
            else:
                pbar.update(len(batch))
                self.db.insert_rows(batch, table_name=schema['table_name'])
        pbar.close()

    def create_schema(self):
        create_table = lambda name, fields: self.db.sql(sql.create_table(name, fields))

        for s in self.schemas:
            create_table(s['table_name'], s['fields'])

    def sql_files(self):
        if 'sql' in self.dataset:
            for f in self.dataset['sql']:
                self.db.execute_sql_file(f)

    def setup_db(self):
        if self.db is None:
            self.db = Database(self.args, table_name=self.name)

    def verify(self):
        self.setup_db()
        verify.check_dataset(self.db, self.name)

    def dump(self):
        """Creates .sql dump file of the datasets"""
        tables = ['--table={}'.format(s['table_name']) for s in self.schemas]
        file_arg = '--file=./{}-{}.sql'.format(self.name, datetime.date.today().isoformat())
        cmd = ["pg_dump", "--no-owner", "--clean", "--if-exists", "-w"] + tables + [file_arg]
        subprocess.run(cmd, env=self.pg_env(), check=True)

    def pg_env(self):
        return merge(
            os.environ.copy(),
            {
                'PGHOST': self.args.host,
                'PGPORT': self.args.port,
                'PGUSER': self.args.user,
                'PGDATABASE': self.args.database,
                'PGPASSWORD': self.args.password
            })



TABLES = {
    'pluto_16v2': {'pluto_16v2': 800000},
    'pluto_17v1': {'pluto_17v1': 800000},
    'pluto_18v1': {'pluto_18v1': 800000},
    'dobjobs': {'dobjobs': 1000000},
    'dob_violations': {'dob_violations': 2000000},
    'ecb_violations': {'ecb_violations': 1300000},
    'hpd_violations': {'hpd_violations': 4000000},
    'hpd_registrations': {
        'hpd_registrations': 150000,
        'hpd_contacts': 600000,
        'hpd_corporate_owners': 30000,
        'hpd_business_addresses': 130000,
        'hpd_registrations_grouped_by_bbl': 130000,
        'hpd_registrations_grouped_by_bbl_with_contacts': 130000
    },
    'dof_sales': {'dof_sales': 75000},
    'rentstab': {'rentstab': 45000},
    'rentstab_summary': {'rentstab_summary': 45000},
    'dob_complaints': {'dob_complaints': 1000000},
    'hpd_complaints': {
        'hpd_complaints': 1000000,
        'hpd_complaint_problems': 2000000
    },
    'acris': {
        'real_property_remarks': 1000000,
        'real_property_legals': 1000000,
        'real_property_master': 1000000,
        'real_property_parties': 1000000,
        'real_property_references': 1000000,
        'personal_property_legals': 1000000,
        'personal_property_master': 1000000,
        'personal_property_parties': 1000000,
        'personal_property_references': 1000000,
        'personal_property_remarks': 400000,
        'acris_country_codes': 250,
        'acris_document_control_codes': 123,
        'acris_property_type_codes': 46,
        'acris_ucc_collateral_codes': 8
    },
    'marshal_evictions_17': {'marshal_evictions_17': 17000},
    'oath_hearings': {'oath_hearings': 10000000}
}


class colors:
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


def check_dataset(db, dataset):
    """ input: nycdb.Database, str """

    for table_name, min_row_count in TABLES[dataset].items():
        if db.table_exists(table_name):
            cnt = db.row_count(table_name)
            if cnt >= min_row_count:
                print(colors.GREEN + table_name + ' has ' + format(cnt, ',') + ' rows' + colors.ENDC)
            else:
                if cnt == 0:
                    print(colors.FAIL + table_name + ' has no rows!' + colors.ENDC)
                else:
                    has_rows = colors.FAIL + table_name + ' has ' + format(cnt, ',') + ' rows.' + colors.ENDC
                    expecting = colors.FAIL + 'Expecting at least ' + colors.BLUE + \
                        format(min_row_count, ',') + colors.ENDC + colors.FAIL + ' rows' + colors.ENDC
                    print(has_rows + expecting)
        else:
            print(colors.FAIL + table_name + ' is missing!' + colors.ENDC)


# 'acris': [
#         'country_codes',
#         'document_control_codes',
#         'ucc_collateral_codes',
#         'personal_property_legals',
#         'personal_property_master',
#         'personal_property_parties',
# #       'personal_property_references',
# #       'personal_property_remarks',
#         'real_property_legals',
#         'real_property_master',
#         'real_property_parties',
#         'real_property_references',
#         'real_property_remarks'
# ],



import os
import subprocess
from tempfile import TemporaryDirectory
from zipfile import ZipFile


def map_pluto(dataset=None, schema=None):
    """
    Insert MapPluto directly into postgres using the command line program: ogr2ogr

    ogr2ogr will need to be installed and avaible on your path
    On debian/ubuntu install with this command: `sudo apt-get install gdal-bin`
    """
    with TemporaryDirectory() as temp_dir:
        with ZipFile(dataset.files[0].dest, 'r') as zip_file:
            zip_file.extractall(path=temp_dir)

            ogr_args = ['ogr2ogr',
                        '-f', 'PostgreSQL',
                        dataset.db.ogr_connection_str(),
                        '-nln', schema['table_name'],
                        '-nlt', 'promote_to_multi',
                        '-lco', 'precision=NO',
                        os.path.join(temp_dir, 'MapPLUTO.shp')]

            subprocess.run(ogr_args, check=True)



import argparse
import logging
import os
import subprocess
import sys
from .dataset import Dataset, datasets


POSTGRES_USER = os.environ.get('NYCDB_POSTGRES_USER', 'nycdb')
POSTGRES_PASSWORD = os.environ.get('NYCDB_POSTGRES_PASSWORD', 'nycdb')
POSTGRES_HOST = os.environ.get('NYCDB_POSTGRES_HOST', '127.0.0.1')
POSTGRES_DB = os.environ.get('NYCDB_POSTGRES_DB', 'nycdb')
POSTGRES_PORT = os.environ.get('NYCDB_POSTGRES_PORT', '5432')


def parse_args():
    parser = argparse.ArgumentParser(description='NYC-DB: utilities for the database of NYC housing data')

    # Download, Load, Verify, Dump
    parser.add_argument('--download', action='store', help='downloads file for provided dataset')
    parser.add_argument('--load', action='store', help='loads dataset into postgres')
    parser.add_argument('--verify', action='store', help='verifies a dataset by checking the table row count')
    parser.add_argument('--dump', action='store', help='creates a sql dump of the datasets in the current folder')
    # list and verify
    parser.add_argument('--list-datasets', action='store_true', help='lists all datasets')
    parser.add_argument('--verify-all', action='store_true', help='verifies all datasets')
    # DB CONNECTION
    parser.add_argument(
        "-U",
        "--user",
        help="Postgres user. default: {}".format(POSTGRES_USER),
        default=POSTGRES_USER
    )
    parser.add_argument(
        "-P",
        "--password",
        help="Postgres password. default: {}".format(POSTGRES_PASSWORD),
        default=POSTGRES_PASSWORD
    )
    parser.add_argument(
        "-H",
        "--host",
        help="Postgres host: default: {}".format(POSTGRES_HOST),
        default=POSTGRES_HOST
    )
    parser.add_argument(
        "-D",
        "--database",
        help="postgres database: default: {}".format(POSTGRES_DB),
        default=POSTGRES_DB
    )
    parser.add_argument(
        "--port",
        help="Postgres port: default: {}".format(POSTGRES_PORT),
        default=POSTGRES_PORT
    )
    # change location of data dir
    parser.add_argument("--root-dir", help="location of data directory", default="./data")
    # easily inspect the database from the command-line
    parser.add_argument("--dbshell", action="store_true", help="runs psql interactively")
    return parser.parse_args()


def print_datasets():
    for ds in datasets().keys():
        print(ds)


def verify_all(args):
    for ds in datasets().keys():
        Dataset(ds, args=args).verify()


def run_dbshell(args):
    env = os.environ.copy()
    env['PGPASSWORD'] = args.password
    retval = subprocess.call([
        'psql', '-h', args.host, '-p', args.port, '-U', args.user, '-d', args.database
    ], env=env)
    sys.exit(retval)


def dispatch(args):
    if args.list_datasets:
        print_datasets()
    elif args.verify:
        Dataset(args.verify, args=args).verify()
    elif args.verify_all:
        verify_all(args)
    elif args.download:
        Dataset(args.download, args=args).download_files()
    elif args.load:
        Dataset(args.load, args=args).db_import()
    elif args.dump:
        Dataset(args.dump, args=args).dump()
    elif args.dbshell:
        run_dbshell(args=args)

def main():
    logging.basicConfig(level=logging.DEBUG)
    args = parse_args()
    dispatch(args)


if __name__ == '__main__':
    main()



""" create BBL  """
import re

manhattan_values = [ 'manhattan', 'mn', '1', 1 ]
bronx_values = [ 'bronx', 'the bronx', 'bx', '2', 2 ]
brooklyn_values = [ 'brooklyn', 'bklyn', 'bk' , '3', 3 ]
queens_values = [ 'queens', 'qn', '4', 4 ]
staten_island_values = [ 'staten island', 'si', '5', 5 ]


def boro_to_code(boro):
    if isinstance(boro, str):
        b = boro.strip().lower()
    else:
        b = boro

    if b in manhattan_values:
        code = '1'
    elif b in bronx_values:
        code = '2'
    elif b in brooklyn_values:
        code = '3'
    elif b in queens_values:
        code = '4'
    elif b in staten_island_values:
        code = '5'
    else:
        # print('Incorrect borough: ' + boro)
        code = '0'
    return code


def lot_length_helper(lot):
    if len(lot) == 5:
        return  lot[1:]
    elif len(lot) < 5:
        return lot.zfill(4)
    else:
        # print(lot + " is more than 5 chars long!")
        return '0000'

def is_empty(i):
    """
    Determines if i is empty
    """
    if isinstance(i, int):
        return False

    if isinstance(i, str) and re.match(r'^\d+$', i.strip()):
        return False

    return True


def bbl(boro, block, lot):
    if boro is None or is_empty(block) or is_empty(lot):
        return None

    return boro_to_code(boro) + str(block).zfill(5) + lot_length_helper(str(lot))
